CON
{{  Notes from
    FullDuplexSerial4portPlus version 1.01
  - Tracy Allen (TTA)   (c)22-Jan-2011   MIT license, see end of file for terms of use.  Extends existing terms of use.
  - Can open up to 4 independent serial ports, using only one pasm cog for all 4.
  - Supports flow control and open and inverted baud modes
  - Individually configurable tx and rx buffers for all 4 ports, any size, set in CONstants section at compile time
  - Part of the buffer fits within the object's hub footprint, but even so the object is restartable
  - Buffers are DAT type variables, therefore a single instance of the object can be accessed throughout a complex project.

  - Modified from Tim Moore's pcFullDuplexSerial4fc, with further motivation and ideas from Duane Degn's pcFullDuplexSerial4fc128
  - Changes and bug fixes include:
    - Flow control is now operational when called for, with correct polarity (bug correction)
    - Jitter is reduced, unused ports are properly skipped over (bug correction), operation speed is increased.
    - Stop bit on reception is now checked, and if there is a framing error, the byte is not put in the buffer.
    - Buffer sizes are arbitrary, each port separate rx & tx up to available memory
       Changes in pasm and in Spin methods to accomodate larger buffers, major reorganization of DAT section.
    - Added strn method for counted string, and rxHowFull method for buffer size.
    - Cut out most of the format methods such as DEC and HEX, expecting those to be their own object calling rx, tx, str and strn methods.
      See companion object DataIO4port.spin in order to maintain compatibility with methods in the original pcFullDuplexSerial4fc.

  - 1v01
    - init method returns pointer @rxsize, for data buffers and data structure.
  - 1v00
    - documentation
  - 0v91
    - restored DEFAULTTHRESHOLD constant
    - made default buffer sizes in the DAT section rather than init
    - removed the numeric methods to their own companion object, dataIO4port.
  - 0v3
    - first public release with the jitter and flow control issues fixed, and large buffers.

  Links:
  Development of this version:
  --- http://forums.parallax.com/showthread.php?137349-yet-another-variant-fullDuplexSerial4portplus
  Tim Moore's original pcFullDuplexSerial4fc and updates to allow flow control:
  --- http://forums.parallaxinc.com/forums/default.aspx?f=25&p=1&m=273291#m276667
  --- http://obex.parallax.com/objects/340/      7/24/08 version
  --- http://forums.parallaxinc.com/forums/default.aspx?f=25&p=1&m=349173    8/14/08 update, flow polarity correction, not in obex
  Duane Degn's thread, larger 128 or 512 byte buffers and reusing buffer space, discussion of issues
  --- http://forums.parallax.com/showthread.php?129714-Tim-Moore-s-pcFullDuplexSerial4FC-with-larger-%28512-byte%29-rx-buffer
  Juergen Buchmueller, 2 port trimmed down version
  --- http://forums.parallax.com/showthread.php?128184-Serial-Objects-for-SPIN-Programming&p=967075&viewfull=1#post967075
  Serial Mirror, single port but same idea regarding buffers in the DATa space
  --- http://forums.parallax.com/showthread.php?94311-SerialMirror-A-FullDuplexSerial-enhancement
  --- http://obex.parallax.com/objects/189/
  Re baud rates attainable, hiccups:
  --- http://forums.parallaxinc.com/forums/default.aspx?f=25&p=1&m=282923#m282978
  --- http://forums.parallaxinc.com/forums/default.aspx?f=25&p=1&m=334784
  --- http://forums.parallax.com/showthread.php?120868-FullDuplexSerial-hiccups
  Jitter, discussions of jitter in different Prop serial programs, PhiPi's development of PBnJ full duplex:
  --- http://forums.parallax.com/showthread.php?129776-Anybody-aware-of-high-accuracy-(0.7-or-less)-serial-full-duplex-driver
  --- http://forums.parallax.com/showthread.php?136431-Serial-objects-question
  Humanoido's catalog of serial port objects
  --- http://forums.parallax.com/showthread.php?128184-Serial-Objects-for-SPIN-Programming

Tim Moore's release notes follow...   Also note by Duane Degn.
Not all these comments apply to  FullDuplexSerial4port.
}}
CON

  NOMODE                        = %000000
  INVERTRX                      = %000001
  INVERTTX                      = %000010
  OCTX                          = %000100
  NOECHO                        = %001000
  INVERTCTS                     = %010000
  INVERTRTS                     = %100000

  PINNOTUSED                    = -1                    'tx/tx/cts/rts pin is not used
  DEFAULTTHRESHOLD              = 0                     ' zero defaults to 3/4 of buffer length

CON

  ' offsets in buffer
  RX_MASK_OFFSET = 0
  TX_MASK_OFFSET = 4
  CTS_MASK_OFFSET = 8
  RTS_MASK_OFFSET = 12
  BIT_4_TICKS_OFFSET = 16
  BIT_TICKS_OFFSET = 20
  LONGS_TO_CLEAR = BIT_TICKS_OFFSET + 4
  
DAT ' Hub Variables

rxchar                  byte 0  ' used by spin rxcheck, for inversion of received data
rxchar1                 byte 0
rxchar2                 byte 0
rxchar3                 byte 0
'cog                     long 0  'cog flag/id
                        org     ' "rxsize" needs to be long aligned.
startFlag               word 0                        
rxsize                  word 0  ' (TTA) size of the rx and tx buffers is available to pasm and spin
rxsize1                 word 0  ' these values are initialized by the "AddPort" method
rxsize2                 word 0  ' at startup, individually configurable
rxsize3                 word 0
txsize                  word 0
txsize1                 word 0
txsize2                 word 0
txsize3                 word 0
rxtx_mode               word 0  ' mode setting from values passed in by addport
rxtx_mode1              word 0  ' used by Spin and copied to PASM           '
rxtx_mode2              word 0
rxtx_mode3              word 0
rxbuff_ptr              word 0  ' These are the base hub addresses of the receive buffers
rxbuff_ptr1             word 0  ' initialized in Spin, referenced in pasm and Spin
rxbuff_ptr2             word 0  ' these buffers and sizes are individually configurable
rxbuff_ptr3             word 0
txbuff_ptr              word 0  ' These are the base hub addresses of the transmit buffers
txbuff_ptr1             word 0  ' Used by "tx" method and in PASM.
txbuff_ptr2             word 0
txbuff_ptr3             word 0

rtssize                 word 0  ' threshold in count of bytes above which will assert rts to stop flow
rtssize1                word 0  ' Used only in PASM
rtssize2                word 0
rtssize3                word 0
rxbuff_head_ptr         word 0  ' Hub address of data received, base plus offset
rxbuff_head_ptr1        word 0  ' pasm writes WRBYTE to hub at this address, initialized in spin to base address
rxbuff_head_ptr2        word 0  ' Used only in PASM
rxbuff_head_ptr3        word 0
txbuff_tail_ptr         word 0  ' Hub address of data tranmitted, base plus offset
txbuff_tail_ptr1        word 0  ' pasm reads RDBYTE from hub at this address, initialized in spin to base address
txbuff_tail_ptr2        word 0  ' Used only in PASM
txbuff_tail_ptr3        word 0


rx_head_ptr             word 0  ' pointer to the hub address of where the head and tail offset pointers are stored
                                 ' these pointers are initialized in spin but then used only by pasm
                                ' the pasm cog has to know where in the hub to find those offsets.

'' The variables below are used only by the Spin section of the program.
bufferPtr               word 0  ' Should point to a long aligned location.
rx_head                 word 0  ' rx head pointer, from 0 to size of rx buffer, used in spin and pasm
rx_head1                word 0  ' data is enqueued to this offset above base, rxbuff_ptr
rx_head2                word 0
rx_head3                word 0
rx_tail                 word 0  ' rx tail pointer, ditto, zero to size of rx buffer
rx_tail1                word 0  ' data is dequeued from this offset above base, rxbuff_ptr
rx_tail2                word 0
rx_tail3                word 0
tx_head                 word 0  ' tx head pointer, , from 0 to size of tx buffer, used in spin only
tx_head1                word 0  ' data is enqueued to this offset above base, txbuff_ptr
tx_head2                word 0
tx_head3                word 0
tx_tail                 word 0  ' tx tail pointer, ditto, zero to size of rx buffer
tx_tail1                word 0  ' data is transmitted from this offset above base, txbuff_ptr
tx_tail2                word 0
tx_tail3                word 0

PUB Init(pasmAddress, bufferAddress)
'' Always call init before adding ports
'' The buffer at location "bufferAddress" should be
'' long aligned. 96 bytes of the buffer will temporarily
'' be used to pass values to PASM. The parent object needs
'' to make sure there are 96 bytes available at the location
'' of "bufferAddress". This will unlikely be a problem since
'' the combined buffers of the serial driver will likely
'' exceed 96 bytes.

  bufferPtr := bufferAddress

  result := cognew(pasmAddress, @startFlag) + 1
  waitcnt(80_000 + cnt) ' give time for PASM to load completely
  longfill(bufferPtr, 0, LONGS_TO_CLEAR) ' unused port's values will be set to zero
                                        
PUB AddPort(port, rxpin, txpin, ctspin, rtspin, rtsthreshold, mode, baudrate, {
} rxBufferSize, txBufferSize)
'' Call AddPort to define each port
'' port 0-3 port index of which serial port
'' rx/tx/cts/rtspin pin number                          XXX#PINNOTUSED if not used
'' rtsthreshold - buffer threshold before rts is used   XXX#DEFAULTTHRSHOLD means use default
'' mode bit 0 = invert rx                               XXX#INVERTRX
'' mode bit 1 = invert tx                               XXX#INVERTTX
'' mode bit 2 = open-drain/source tx                    XXX#OCTX
'' mode bit 3 = ignore tx echo on rx                    XXX#NOECHO
'' mode bit 4 = invert cts                              XXX#INVERTCTS
'' mode bit 5 = invert rts                              XXX#INVERTRTS
'' baudrate

  if rxpin <> -1
    long[bufferPtr][RX_MASK_OFFSET + port] := |< rxpin
  if txpin <> -1
    long[bufferPtr][TX_MASK_OFFSET + port] := |< txpin
  if ctspin <> -1
    long[bufferPtr][CTS_MASK_OFFSET + port] := |< ctspin
  if rtspin <> -1
    long[bufferPtr][RTS_MASK_OFFSET + port] := |< rtspin 
    if (rtsthreshold > 0) and (rtsthreshold < rxsize[port])           ' (TTA) modified for variable buffer size
      rtssize[port] := rtsthreshold
    else
      rtssize[port] := rxsize[port]*3/4                        'default rts threshold 3/4 of buffer  TTS ref RX_BUFSIZE
  rxtx_mode[port] := mode
  if mode & INVERTRX
   rxchar[port] := $ff
  long[bufferPtr][BIT_TICKS_OFFSET + port] := (clkfreq / baudrate)
  long[bufferPtr][BIT_4_TICKS_OFFSET + port] := long[bufferPtr][BIT_TICKS_OFFSET + port] >> 2
  rxsize[port] := rxBufferSize
  txsize[port] := txBufferSize

PUB Start 
'' Call start to allow previously starter cog to execute PASM code.
'' Start serial driver 


  txbuff_tail_ptr := txbuff_ptr  := bufferPtr                 ' (TTA) all buffers are calculated as offsets from this address.
  txbuff_tail_ptr1 := txbuff_ptr1 := txbuff_ptr + txsize      'base addresses of the corresponding port buffer.
  txbuff_tail_ptr2 := txbuff_ptr2 := txbuff_ptr1 + txsize1
  txbuff_tail_ptr3 := txbuff_ptr3 := txbuff_ptr2 + txsize2
  rxbuff_head_ptr := rxbuff_ptr := txbuff_ptr3 + txsize3     ' rx buffers follow immediately after the tx buffers, by size
  rxbuff_head_ptr1 := rxbuff_ptr1 := rxbuff_ptr + rxsize
  rxbuff_head_ptr2 := rxbuff_ptr2 := rxbuff_ptr1 + rxsize1
  rxbuff_head_ptr3 := rxbuff_ptr3 := rxbuff_ptr2 + rxsize2
                                                        ' note that txbuff_ptr ... rxbuff_ptr3 are the base addresses fixed
                                                        ' in memory for use by both spin and pasm
                                                        ' while txbuff_tail_ptr ... rxbuff_head_ptr3 are dynamic addresses used only by pasm
                                                        ' and here initialized to point to the start of the buffers.
                                                             ' the rx buffer #3 comes last, up through address @endfill
  rx_head_ptr := @rx_head                               ' (TTA) note: addresses of the head and tail counts are passed to the cog
                                                        ' if that is confusing, take heart.   These are pointers to pointers to pointers
  startFlag := 1
  
{PUB GetFirst

  result := @entry

PUB GetLast

  result := @entry
}
PUB rxflush(port)
'' Flush receive buffer, here until empty.
  repeat while rxcheck(port) => 0

PUB rxHowFull(port)    ' (TTA) added method
'' returns number of chars in rx buffer
  return ((rx_head[port] - rx_tail[port]) + rxsize[port]) // rxsize[port]
'   rx_head and rx_tail are values in the range 0=< ... < RX_BUFSIZE


PUB rxcheck(port) : rxbyte
'' Check if byte received (never waits)
'' returns -1 if no byte received, $00..$FF if byte
'' (TTA) simplified references
  if port > 3
    abort
  rxbyte--
  if rx_tail[port] <> rx_head[port]
    rxbyte := rxchar[port] ^ byte[rxbuff_ptr[port]+rx_tail[port]]
    rx_tail[port] := (rx_tail[port] + 1) // rxsize[port]

PUB rxtime(port,ms) : rxbyte | t
'' Wait ms milliseconds for a byte to be received
'' returns -1 if no byte received, $00..$FF if byte
  t := cnt
  repeat until (rxbyte := rxcheck(port)) => 0 or (cnt - t) / (clkfreq / 1000) > ms

PUB rx(port) : rxbyte
'' Receive byte (may wait for byte)
'' returns $00..$FF
  repeat while (rxbyte := rxcheck(port)) < 0

PUB tx(port,txbyte)
'' Send byte (may wait for room in buffer)
  if port > 3
    abort
  repeat until (tx_tail[port] <> (tx_head[port] + 1) // txsize[port])
  byte[txbuff_ptr[port]+tx_head[port]] := txbyte
  tx_head[port] := (tx_head[port] + 1) // txsize[port]

  if rxtx_mode[port] & NOECHO
    rx(port)

{PUB txflush(port)
  repeat until (long[@tx_tail][port] == long[@tx_head][port])
  }
PUB str(port,stringptr)
'' Send zstring
  strn(port,stringptr,strsize(stringptr))

PUB strn(port,stringptr,nchar)
'' Send counted string
  repeat nchar
    tx(port,byte[stringptr++])
 
DAT

{{
+------------------------------------------------------------------------------------------------------------------------------+
|                                                   TERMS OF USE: MIT License                                                  |                                                            
+------------------------------------------------------------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    | 
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
+------------------------------------------------------------------------------------------------------------------------------+
}}